# Workaround for good IDE and universal for runtime
# noinspection PyUnreachableCode
if False:
    from ._grpc.v4.protos import ydb_scripting_pb2
    from ._grpc.v4 import ydb_scripting_v1_pb2_grpc
else:
    from ._grpc.common.protos import ydb_scripting_pb2
    from ._grpc.common import ydb_scripting_v1_pb2_grpc


from . import issues, convert, settings


class TypedParameters(object):
    def __init__(self, parameters_types, parameters_values):
        self.parameters_types = parameters_types
        self.parameters_values = parameters_values


class ScriptingClientSettings(object):
    def __init__(self):
        self._native_date_in_result_sets = False
        self._native_datetime_in_result_sets = False

    def with_native_date_in_result_sets(self, enabled):
        self._native_date_in_result_sets = enabled
        return self

    def with_native_datetime_in_result_sets(self, enabled):
        self._native_datetime_in_result_sets = enabled
        return self


class ExplainYqlScriptSettings(settings.BaseRequestSettings):
    MODE_UNSPECIFIED = 0
    MODE_PARSE = 1
    MODE_VALIDATE = 2
    MODE_EXPLAIN = 3

    def __init__(self):
        super(ExplainYqlScriptSettings, self).__init__()
        self.mode = False

    def with_mode(self, val):
        self.mode = val
        return self


def _execute_yql_query_request_factory(script, tp=None, settings=None):
    params = (
        None
        if tp is None
        else convert.parameters_to_pb(tp.parameters_types, tp.parameters_values)
    )
    return ydb_scripting_pb2.ExecuteYqlRequest(script=script, parameters=params)


class YqlQueryResult(object):
    def __init__(self, result, scripting_client_settings=None):
        self.result_sets = convert.ResultSets(
            result.result_sets, scripting_client_settings
        )


class YqlExplainResult(object):
    def __init__(self, result):
        self.plan = result.plan


def _wrap_response(rpc_state, response, scripting_client_settings):
    issues._process_response(response.operation)
    message = ydb_scripting_pb2.ExecuteYqlResult()
    response.operation.result.Unpack(message)
    return YqlQueryResult(message)


def _wrap_explain_response(rpc_state, response):
    issues._process_response(response.operation)
    message = ydb_scripting_pb2.ExplainYqlResult()
    response.operation.result.Unpack(message)
    return YqlExplainResult(message)


class ScriptingClient(object):
    def __init__(self, driver, scripting_client_settings=None):
        self.driver = driver
        self.scripting_client_settings = (
            scripting_client_settings
            if scripting_client_settings is not None
            else ScriptingClientSettings()
        )

    def execute_yql(self, script, typed_parameters=None, settings=None):
        request = _execute_yql_query_request_factory(script, typed_parameters, settings)
        return self.driver(
            request,
            ydb_scripting_v1_pb2_grpc.ScriptingServiceStub,
            "ExecuteYql",
            _wrap_response,
            settings=settings,
            wrap_args=(self.scripting_client_settings,),
        )

    def explain_yql(self, script, settings=None):
        return self.driver(
            ydb_scripting_pb2.ExplainYqlRequest(script=script, mode=settings.mode),
            ydb_scripting_v1_pb2_grpc.ScriptingServiceStub,
            "ExplainYql",
            _wrap_explain_response,
            settings=settings,
        )



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from ydb._grpc.v4.protos import ydb_coordination_pb2 as protos_dot_ydb__coordination__pb2


class CoordinationServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Session = channel.stream_stream(
                '/Ydb.Coordination.V1.CoordinationService/Session',
                request_serializer=protos_dot_ydb__coordination__pb2.SessionRequest.SerializeToString,
                response_deserializer=protos_dot_ydb__coordination__pb2.SessionResponse.FromString,
                )
        self.CreateNode = channel.unary_unary(
                '/Ydb.Coordination.V1.CoordinationService/CreateNode',
                request_serializer=protos_dot_ydb__coordination__pb2.CreateNodeRequest.SerializeToString,
                response_deserializer=protos_dot_ydb__coordination__pb2.CreateNodeResponse.FromString,
                )
        self.AlterNode = channel.unary_unary(
                '/Ydb.Coordination.V1.CoordinationService/AlterNode',
                request_serializer=protos_dot_ydb__coordination__pb2.AlterNodeRequest.SerializeToString,
                response_deserializer=protos_dot_ydb__coordination__pb2.AlterNodeResponse.FromString,
                )
        self.DropNode = channel.unary_unary(
                '/Ydb.Coordination.V1.CoordinationService/DropNode',
                request_serializer=protos_dot_ydb__coordination__pb2.DropNodeRequest.SerializeToString,
                response_deserializer=protos_dot_ydb__coordination__pb2.DropNodeResponse.FromString,
                )
        self.DescribeNode = channel.unary_unary(
                '/Ydb.Coordination.V1.CoordinationService/DescribeNode',
                request_serializer=protos_dot_ydb__coordination__pb2.DescribeNodeRequest.SerializeToString,
                response_deserializer=protos_dot_ydb__coordination__pb2.DescribeNodeResponse.FromString,
                )


class CoordinationServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def Session(self, request_iterator, context):
        """*
        Bidirectional stream used to establish a session with a coordination node

        Relevant APIs for managing semaphores, distributed locking, creating or
        restoring a previously established session are described using nested
        messages in SessionRequest and SessionResponse. Session is established
        with a specific coordination node (previously created using CreateNode
        below) and semaphores are local to that coordination node.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CreateNode(self, request, context):
        """Creates a new coordination node
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def AlterNode(self, request, context):
        """Modifies settings of a coordination node
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DropNode(self, request, context):
        """Drops a coordination node
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DescribeNode(self, request, context):
        """Describes a coordination node
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_CoordinationServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'Session': grpc.stream_stream_rpc_method_handler(
                    servicer.Session,
                    request_deserializer=protos_dot_ydb__coordination__pb2.SessionRequest.FromString,
                    response_serializer=protos_dot_ydb__coordination__pb2.SessionResponse.SerializeToString,
            ),
            'CreateNode': grpc.unary_unary_rpc_method_handler(
                    servicer.CreateNode,
                    request_deserializer=protos_dot_ydb__coordination__pb2.CreateNodeRequest.FromString,
                    response_serializer=protos_dot_ydb__coordination__pb2.CreateNodeResponse.SerializeToString,
            ),
            'AlterNode': grpc.unary_unary_rpc_method_handler(
                    servicer.AlterNode,
                    request_deserializer=protos_dot_ydb__coordination__pb2.AlterNodeRequest.FromString,
                    response_serializer=protos_dot_ydb__coordination__pb2.AlterNodeResponse.SerializeToString,
            ),
            'DropNode': grpc.unary_unary_rpc_method_handler(
                    servicer.DropNode,
                    request_deserializer=protos_dot_ydb__coordination__pb2.DropNodeRequest.FromString,
                    response_serializer=protos_dot_ydb__coordination__pb2.DropNodeResponse.SerializeToString,
            ),
            'DescribeNode': grpc.unary_unary_rpc_method_handler(
                    servicer.DescribeNode,
                    request_deserializer=protos_dot_ydb__coordination__pb2.DescribeNodeRequest.FromString,
                    response_serializer=protos_dot_ydb__coordination__pb2.DescribeNodeResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'Ydb.Coordination.V1.CoordinationService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class CoordinationService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def Session(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_stream(request_iterator, target, '/Ydb.Coordination.V1.CoordinationService/Session',
            protos_dot_ydb__coordination__pb2.SessionRequest.SerializeToString,
            protos_dot_ydb__coordination__pb2.SessionResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def CreateNode(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Ydb.Coordination.V1.CoordinationService/CreateNode',
            protos_dot_ydb__coordination__pb2.CreateNodeRequest.SerializeToString,
            protos_dot_ydb__coordination__pb2.CreateNodeResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def AlterNode(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Ydb.Coordination.V1.CoordinationService/AlterNode',
            protos_dot_ydb__coordination__pb2.AlterNodeRequest.SerializeToString,
            protos_dot_ydb__coordination__pb2.AlterNodeResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def DropNode(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Ydb.Coordination.V1.CoordinationService/DropNode',
            protos_dot_ydb__coordination__pb2.DropNodeRequest.SerializeToString,
            protos_dot_ydb__coordination__pb2.DropNodeResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def DescribeNode(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Ydb.Coordination.V1.CoordinationService/DescribeNode',
            protos_dot_ydb__coordination__pb2.DescribeNodeRequest.SerializeToString,
            protos_dot_ydb__coordination__pb2.DescribeNodeResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



from unittest import mock
from ydb import (
    retry_operation_impl,
    YdbRetryOperationFinalResult,
    issues,
    YdbRetryOperationSleepOpt,
    RetrySettings,
)


def test_retry_operation_impl(monkeypatch):
    monkeypatch.setattr("random.random", lambda: 0.5)
    monkeypatch.setattr(
        issues.Error,
        "__eq__",
        lambda self, other: type(self) == type(other) and self.message == other.message,
    )

    retry_once_settings = RetrySettings(
        max_retries=1,
        on_ydb_error_callback=mock.Mock(),
    )
    retry_once_settings.unknown_error_handler = mock.Mock()

    def get_results(callee):
        res_generator = retry_operation_impl(callee, retry_settings=retry_once_settings)
        results = []
        exc = None
        try:
            for res in res_generator:
                results.append(res)
                if isinstance(res, YdbRetryOperationFinalResult):
                    break
        except Exception as e:
            exc = e

        return results, exc

    class TestException(Exception):
        def __init__(self, message):
            super(TestException, self).__init__(message)
            self.message = message

        def __eq__(self, other):
            return type(self) == type(other) and self.message == other.message

    def check_unretriable_error(err_type, call_ydb_handler):
        retry_once_settings.on_ydb_error_callback.reset_mock()
        retry_once_settings.unknown_error_handler.reset_mock()

        results = get_results(
            mock.Mock(side_effect=[err_type("test1"), err_type("test2")])
        )
        yields = results[0]
        exc = results[1]

        assert yields == []
        assert exc == err_type("test1")

        if call_ydb_handler:
            assert retry_once_settings.on_ydb_error_callback.call_count == 1
            retry_once_settings.on_ydb_error_callback.assert_called_with(
                err_type("test1")
            )

            assert retry_once_settings.unknown_error_handler.call_count == 0
        else:
            assert retry_once_settings.on_ydb_error_callback.call_count == 0

            assert retry_once_settings.unknown_error_handler.call_count == 1
            retry_once_settings.unknown_error_handler.assert_called_with(
                err_type("test1")
            )

    def check_retriable_error(err_type, backoff):
        retry_once_settings.on_ydb_error_callback.reset_mock()

        results = get_results(
            mock.Mock(side_effect=[err_type("test1"), err_type("test2")])
        )
        yields = results[0]
        exc = results[1]

        if backoff:
            assert [
                YdbRetryOperationSleepOpt(backoff.calc_timeout(0)),
                YdbRetryOperationSleepOpt(backoff.calc_timeout(1)),
            ] == yields
        else:
            assert [] == yields

        assert exc == err_type("test2")

        assert retry_once_settings.on_ydb_error_callback.call_count == 2
        retry_once_settings.on_ydb_error_callback.assert_any_call(err_type("test1"))
        retry_once_settings.on_ydb_error_callback.assert_called_with(err_type("test2"))

        assert retry_once_settings.unknown_error_handler.call_count == 0

    # check ok
    assert get_results(lambda: True) == ([YdbRetryOperationFinalResult(True)], None)

    # check retry error and return result
    assert get_results(mock.Mock(side_effect=[issues.Overloaded("test"), True])) == (
        [
            YdbRetryOperationSleepOpt(retry_once_settings.slow_backoff.calc_timeout(0)),
            YdbRetryOperationFinalResult(True),
        ],
        None,
    )

    # check errors
    check_retriable_error(issues.Aborted, None)
    check_retriable_error(issues.BadSession, None)

    check_retriable_error(issues.NotFound, None)
    with mock.patch.object(retry_once_settings, "retry_not_found", False):
        check_unretriable_error(issues.NotFound, True)

    check_retriable_error(issues.InternalError, None)
    with mock.patch.object(retry_once_settings, "retry_internal_error", False):
        check_unretriable_error(issues.InternalError, True)

    check_retriable_error(issues.Overloaded, retry_once_settings.slow_backoff)
    check_retriable_error(issues.SessionPoolEmpty, retry_once_settings.slow_backoff)
    check_retriable_error(issues.ConnectionError, retry_once_settings.slow_backoff)

    check_retriable_error(issues.Unavailable, retry_once_settings.fast_backoff)

    check_unretriable_error(issues.Undetermined, True)
    with mock.patch.object(retry_once_settings, "idempotent", True):
        check_retriable_error(issues.Unavailable, retry_once_settings.fast_backoff)

    check_unretriable_error(issues.Error, True)
    with mock.patch.object(retry_once_settings, "idempotent", True):
        check_unretriable_error(issues.Error, True)

    check_unretriable_error(TestException, False)
    with mock.patch.object(retry_once_settings, "idempotent", True):
        check_unretriable_error(TestException, False)



# -*- coding: utf-8 -*-
import setuptools

with open("README.md", "r") as r:
    long_description = r.read()

with open("requirements.txt") as r:
    requirements = []
    for line in r.readlines():
        line = line.strip()
        if line != "":
            requirements.append(line)

setuptools.setup(
    name="ydb",
    version="2.13.1",  # AUTOVERSION
    description="YDB Python SDK",
    author="Yandex LLC",
    author_email="ydb@yandex-team.ru",
    url="http://github.com/ydb-platform/ydb-python-sdk",
    license="Apache 2.0",
    package_dir={"": "."},
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=setuptools.find_packages("."),
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
    ],
    install_requires=requirements,  # requirements.txt
    options={"bdist_wheel": {"universal": True}},
    extras_require={
        "yc": ["yandexcloud", ],
    }
)



# -*- coding: utf-8 -*-
# flake8: noqa
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ydb_scripting_v1.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ydb._grpc.v3.protos import ydb_scripting_pb2 as protos_dot_ydb__scripting__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='ydb_scripting_v1.proto',
  package='Ydb.Scripting.V1',
  syntax='proto3',
  serialized_options=b'\n\025tech.ydb.scripting.v1Z8github.com/ydb-platform/ydb-go-genproto/Ydb_Scripting_V1',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x16ydb_scripting_v1.proto\x12\x10Ydb.Scripting.V1\x1a\x1aprotos/ydb_scripting.proto2\x9a\x02\n\x10ScriptingService\x12Q\n\nExecuteYql\x12 .Ydb.Scripting.ExecuteYqlRequest\x1a!.Ydb.Scripting.ExecuteYqlResponse\x12`\n\x10StreamExecuteYql\x12 .Ydb.Scripting.ExecuteYqlRequest\x1a(.Ydb.Scripting.ExecuteYqlPartialResponse0\x01\x12Q\n\nExplainYql\x12 .Ydb.Scripting.ExplainYqlRequest\x1a!.Ydb.Scripting.ExplainYqlResponseBQ\n\x15tech.ydb.scripting.v1Z8github.com/ydb-platform/ydb-go-genproto/Ydb_Scripting_V1b\x06proto3'
  ,
  dependencies=[protos_dot_ydb__scripting__pb2.DESCRIPTOR,])



_sym_db.RegisterFileDescriptor(DESCRIPTOR)


DESCRIPTOR._options = None

_SCRIPTINGSERVICE = _descriptor.ServiceDescriptor(
  name='ScriptingService',
  full_name='Ydb.Scripting.V1.ScriptingService',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=73,
  serialized_end=355,
  methods=[
  _descriptor.MethodDescriptor(
    name='ExecuteYql',
    full_name='Ydb.Scripting.V1.ScriptingService.ExecuteYql',
    index=0,
    containing_service=None,
    input_type=protos_dot_ydb__scripting__pb2._EXECUTEYQLREQUEST,
    output_type=protos_dot_ydb__scripting__pb2._EXECUTEYQLRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='StreamExecuteYql',
    full_name='Ydb.Scripting.V1.ScriptingService.StreamExecuteYql',
    index=1,
    containing_service=None,
    input_type=protos_dot_ydb__scripting__pb2._EXECUTEYQLREQUEST,
    output_type=protos_dot_ydb__scripting__pb2._EXECUTEYQLPARTIALRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ExplainYql',
    full_name='Ydb.Scripting.V1.ScriptingService.ExplainYql',
    index=2,
    containing_service=None,
    input_type=protos_dot_ydb__scripting__pb2._EXPLAINYQLREQUEST,
    output_type=protos_dot_ydb__scripting__pb2._EXPLAINYQLRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_SCRIPTINGSERVICE)

DESCRIPTOR.services_by_name['ScriptingService'] = _SCRIPTINGSERVICE

# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
import threading
import logging
from concurrent import futures
import collections
import random

import six

from . import connection as connection_impl, issues, resolver, _utilities, tracing
from abc import abstractmethod, ABCMeta


logger = logging.getLogger(__name__)


class ConnectionsCache(object):
    def __init__(self, use_all_nodes=False, tracer=tracing.Tracer(None)):
        self.tracer = tracer
        self.lock = threading.RLock()
        self.connections = collections.OrderedDict()
        self.connections_by_node_id = collections.OrderedDict()
        self.outdated = collections.OrderedDict()
        self.subscriptions = set()
        self.preferred = collections.OrderedDict()
        self.logger = logging.getLogger(__name__)
        self.use_all_nodes = use_all_nodes
        self.conn_lst_order = (
            (self.connections,)
            if self.use_all_nodes
            else (self.preferred, self.connections)
        )
        self.fast_fail_subscriptions = set()

    def add(self, connection, preferred=False):
        if connection is None:
            return False

        connection.add_cleanup_callback(self.remove)
        with self.lock:
            if preferred:
                self.preferred[connection.endpoint] = connection

            self.connections_by_node_id[connection.node_id] = connection
            self.connections[connection.endpoint] = connection
            subscriptions = list(self.subscriptions)
            self.subscriptions.clear()

            if len(self.connections) > 0:
                self.complete_discovery(None)

        for subscription in subscriptions:
            subscription.set_result(None)
        return True

    def _on_done_callback(self, subscription):
        """
        A done callback for the subscription future
        :param subscription: A subscription
        :return: None
        """
        with self.lock:
            try:
                self.subscriptions.remove(subscription)
            except KeyError:
                return subscription

    @property
    def size(self):
        with self.lock:
            return len(self.connections) - len(self.outdated)

    def already_exists(self, endpoint):
        with self.lock:
            return endpoint in self.connections

    def values(self):
        with self.lock:
            return list(self.connections.values())

    def make_outdated(self, connection):
        with self.lock:
            self.outdated[connection.endpoint] = connection
            return self

    def cleanup_outdated(self):
        with self.lock:
            outdated_connections = list(self.outdated.values())
            for outdated_connection in outdated_connections:
                outdated_connection.close()
        return self

    def cleanup(self):
        with self.lock:
            actual_connections = list(self.connections.values())
            for connection in actual_connections:
                connection.close()

    def complete_discovery(self, error):
        with self.lock:
            for subscription in self.fast_fail_subscriptions:
                if error is None:
                    subscription.set_result(None)
                else:
                    subscription.set_exception(error)

            self.fast_fail_subscriptions.clear()

    def add_fast_fail(self):
        with self.lock:
            subscription = futures.Future()
            if len(self.connections) > 0:
                subscription.set_result(None)
                return subscription

            self.fast_fail_subscriptions.add(subscription)
            return subscription

    def subscribe(self):
        with self.lock:
            subscription = futures.Future()
            if len(self.connections) > 0:
                subscription.set_result(None)
                return subscription
            self.subscriptions.add(subscription)
            subscription.add_done_callback(self._on_done_callback)
            return subscription

    @tracing.with_trace()
    def get(self, preferred_endpoint=None):
        with self.lock:
            if (
                preferred_endpoint is not None
                and preferred_endpoint.node_id in self.connections_by_node_id
            ):
                return self.connections_by_node_id[preferred_endpoint.node_id]

            if (
                preferred_endpoint is not None
                and preferred_endpoint.endpoint in self.connections
            ):
                return self.connections[preferred_endpoint]

            for conn_lst in self.conn_lst_order:
                try:
                    endpoint, connection = conn_lst.popitem(last=False)
                    conn_lst[endpoint] = connection
                    tracing.trace(self.tracer, {"found_in_lists": True})
                    return connection
                except KeyError:
                    continue

            raise issues.ConnectionLost("Couldn't find valid connection")

    def remove(self, connection):
        with self.lock:
            self.connections_by_node_id.pop(connection.node_id, None)
            self.preferred.pop(connection.endpoint, None)
            self.connections.pop(connection.endpoint, None)
            self.outdated.pop(connection.endpoint, None)


class Discovery(threading.Thread):
    def __init__(self, store, driver_config):
        """
        A timer thread that implements endpoints discovery logic
        :param store: A store with endpoints
        :param driver_config: An instance of DriverConfig
        """
        super(Discovery, self).__init__()
        self.logger = logger.getChild(self.__class__.__name__)
        self.condition = threading.Condition()
        self.daemon = True
        self._cache = store
        self._driver_config = driver_config
        self._resolver = resolver.DiscoveryEndpointsResolver(self._driver_config)
        self._base_discovery_interval = 60
        self._ready_timeout = 4
        self._discovery_request_timeout = 2
        self._should_stop = threading.Event()
        self._max_size = 9
        self._base_emergency_retry_interval = 1
        self._ssl_required = False
        if driver_config.root_certificates is not None or driver_config.secure_channel:
            self._ssl_required = True

    def discovery_debug_details(self):
        return self._resolver.debug_details()

    def _emergency_retry_interval(self):
        return (1 + random.random()) * self._base_emergency_retry_interval

    def _discovery_interval(self):
        return (1 + random.random()) * self._base_discovery_interval

    def notify_disconnected(self):
        self._send_wake_up()

    def _send_wake_up(self):
        acquired = self.condition.acquire(blocking=False)

        if not acquired:
            return

        self.condition.notify_all()
        self.condition.release()

    def _handle_empty_database(self):
        if self._cache.size > 0:
            return True

        return self._cache.add(
            connection_impl.Connection.ready_factory(
                self._driver_config.endpoint, self._driver_config, self._ready_timeout
            )
        )

    def execute_discovery(self):
        if self._driver_config.database is None:
            return self._handle_empty_database()

        with self._resolver.context_resolve() as resolve_details:
            if resolve_details is None:
                return False

            resolved_endpoints = set(
                endpoint
                for resolved_endpoint in resolve_details.endpoints
                for endpoint, endpoint_options in resolved_endpoint.endpoints_with_options()
            )
            for cached_endpoint in self._cache.values():
                if cached_endpoint.endpoint not in resolved_endpoints:
                    self._cache.make_outdated(cached_endpoint)

            for resolved_endpoint in resolve_details.endpoints:
                if self._ssl_required and not resolved_endpoint.ssl:
                    continue

                if not self._ssl_required and resolved_endpoint.ssl:
                    continue

                preferred = resolve_details.self_location == resolved_endpoint.location

                for (
                    endpoint,
                    endpoint_options,
                ) in resolved_endpoint.endpoints_with_options():
                    if (
                        self._cache.size >= self._max_size
                        or self._cache.already_exists(endpoint)
                    ):
                        continue

                    ready_connection = connection_impl.Connection.ready_factory(
                        endpoint,
                        self._driver_config,
                        self._ready_timeout,
                        endpoint_options=endpoint_options,
                    )
                    self._cache.add(ready_connection, preferred)

        self._cache.cleanup_outdated()

        return self._cache.size > 0

    def stop(self):
        self._should_stop.set()
        self._send_wake_up()

    def run(self):
        with self.condition:
            while True:
                if self._should_stop.is_set():
                    break

                successful = self.execute_discovery()
                if successful:
                    self._cache.complete_discovery(None)
                else:
                    self._cache.complete_discovery(
                        issues.ConnectionFailure(str(self.discovery_debug_details()))
                    )

                if self._should_stop.is_set():
                    break

                interval = (
                    self._discovery_interval()
                    if successful
                    else self._emergency_retry_interval()
                )
                self.condition.wait(interval)

            self._cache.cleanup()
        self.logger.info("Successfully terminated discovery process")


@six.add_metaclass(ABCMeta)
class IConnectionPool:
    @abstractmethod
    def __init__(self, driver_config):
        """
        An object that encapsulates discovery logic and provides ability to execute user requests
        on discovered endpoints.
        :param driver_config: An instance of DriverConfig
        """
        pass

    @abstractmethod
    def stop(self, timeout=10):
        """
        Stops underlying discovery process and cleanups
        :param timeout: A timeout to wait for stop completion
        :return: None
        """
        pass

    @abstractmethod
    def wait(self, timeout=None, fail_fast=False):
        """
        Waits for endpoints to be are available to serve user requests
        :param timeout: A timeout to wait in seconds
        :param fail_fast: Should wait fail fast?
        :return: None
        """

    @abstractmethod
    def discovery_debug_details(self):
        """
        Returns debug string about last errors
        :return:
        """
        pass

    @abstractmethod
    def __call__(
        self,
        request,
        stub,
        rpc_name,
        wrap_result=None,
        settings=None,
        wrap_args=(),
        preferred_endpoint=None,
    ):
        """
        Sends request constructed by client library
        :param request: A request constructed by client
        :param stub: A stub instance to wrap channel
        :param rpc_name: A name of RPC to be executed
        :param wrap_result: A callable that intercepts call and wraps received response
        :param settings: An instance of BaseRequestSettings that can be used
        for RPC metadata construction
        :param wrap_args: And arguments to be passed into wrap_result callable
        :return: A result of computation
        """
        pass


class ConnectionPool(IConnectionPool):
    def __init__(self, driver_config):
        """
        An object that encapsulates discovery logic and provides ability to execute user requests
        on discovered endpoints.

        :param driver_config: An instance of DriverConfig
        """
        self._driver_config = driver_config
        self._store = ConnectionsCache(
            driver_config.use_all_nodes, driver_config.tracer
        )
        self.tracer = driver_config.tracer
        self._grpc_init = connection_impl.Connection(
            self._driver_config.endpoint, self._driver_config
        )
        self._discovery_thread = Discovery(self._store, self._driver_config)
        self._discovery_thread.start()
        self._stopped = False
        self._stop_guard = threading.Lock()

    def stop(self, timeout=10):
        """
        Stops underlying discovery process and cleanups

        :param timeout: A timeout to wait for stop completion
        :return: None
        """
        with self._stop_guard:
            if self._stopped:
                return

            self._stopped = True
            self._discovery_thread.stop()
        self._grpc_init.close()
        self._discovery_thread.join(timeout)

    def async_wait(self, fail_fast=False):
        """
        Returns a future to subscribe on endpoints availability.

        :return: A concurrent.futures.Future instance.
        """
        if fail_fast:
            return self._store.add_fast_fail()
        return self._store.subscribe()

    def wait(self, timeout=None, fail_fast=False):
        """
        Waits for endpoints to be are available to serve user requests

        :param timeout: A timeout to wait in seconds
        :return: None
        """
        if fail_fast:
            self._store.add_fast_fail().result(timeout)
        else:
            self._store.subscribe().result(timeout)

    def _on_disconnected(self, connection):
        """
        Removes bad discovered endpoint and triggers discovery process

        :param connection: A disconnected connection
        :return: None
        """
        connection.close()
        self._discovery_thread.notify_disconnected()

    def discovery_debug_details(self):
        return self._discovery_thread.discovery_debug_details()

    @tracing.with_trace()
    def __call__(
        self,
        request,
        stub,
        rpc_name,
        wrap_result=None,
        settings=None,
        wrap_args=(),
        preferred_endpoint=None,
    ):
        """
        Synchronously sends request constructed by client library

        :param request: A request constructed by client
        :param stub: A stub instance to wrap channel
        :param rpc_name: A name of RPC to be executed
        :param wrap_result: A callable that intercepts call and wraps received response
        :param settings: An instance of BaseRequestSettings that can be used
        for RPC metadata construction
        :param wrap_args: And arguments to be passed into wrap_result callable

        :return: A result of computation
        """
        tracing.trace(
            self.tracer, {"request": request, "stub": stub, "rpc_name": rpc_name}
        )
        try:
            connection = self._store.get(preferred_endpoint)
        except Exception:
            self._discovery_thread.notify_disconnected()
            raise

        res = connection(
            request,
            stub,
            rpc_name,
            wrap_result,
            settings,
            wrap_args,
            lambda: self._on_disconnected(connection),
        )
        tracing.trace(
            self.tracer, {"response": res}, trace_level=tracing.TraceLevel.DEBUG
        )
        return res

    @_utilities.wrap_async_call_exceptions
    def future(
        self,
        request,
        stub,
        rpc_name,
        wrap_result=None,
        settings=None,
        wrap_args=(),
        preferred_endpoint=None,
    ):
        """
        Sends request constructed by client

        :param request: A request constructed by client
        :param stub: A stub instance to wrap channel
        :param rpc_name: A name of RPC to be executed
        :param wrap_result: A callable that intercepts call and wraps received response
        :param settings: An instance of BaseRequestSettings that can be used\
        for RPC metadata construction
        :param wrap_args: And arguments to be passed into wrap_result callable

        :return: A future of computation
        """
        try:
            connection = self._store.get(preferred_endpoint)
        except Exception:
            self._discovery_thread.notify_disconnected()
            raise

        return connection.future(
            request,
            stub,
            rpc_name,
            wrap_result,
            settings,
            wrap_args,
            lambda: self._on_disconnected(connection),
        )

    def __enter__(self):
        """
        In some cases (scripts, for example) this context manager can be used.

        :return:
        """
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()



# -*- coding: utf-8 -*-
# flake8: noqa
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ydb_discovery_v1.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ydb._grpc.v3.protos import ydb_discovery_pb2 as protos_dot_ydb__discovery__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='ydb_discovery_v1.proto',
  package='Ydb.Discovery.V1',
  syntax='proto3',
  serialized_options=b'\n\025tech.ydb.discovery.v1Z8github.com/ydb-platform/ydb-go-genproto/Ydb_Discovery_V1',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x16ydb_discovery_v1.proto\x12\x10Ydb.Discovery.V1\x1a\x1aprotos/ydb_discovery.proto2\xb5\x01\n\x10\x44iscoveryService\x12Z\n\rListEndpoints\x12#.Ydb.Discovery.ListEndpointsRequest\x1a$.Ydb.Discovery.ListEndpointsResponse\x12\x45\n\x06WhoAmI\x12\x1c.Ydb.Discovery.WhoAmIRequest\x1a\x1d.Ydb.Discovery.WhoAmIResponseBQ\n\x15tech.ydb.discovery.v1Z8github.com/ydb-platform/ydb-go-genproto/Ydb_Discovery_V1b\x06proto3'
  ,
  dependencies=[protos_dot_ydb__discovery__pb2.DESCRIPTOR,])



_sym_db.RegisterFileDescriptor(DESCRIPTOR)


DESCRIPTOR._options = None

_DISCOVERYSERVICE = _descriptor.ServiceDescriptor(
  name='DiscoveryService',
  full_name='Ydb.Discovery.V1.DiscoveryService',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=73,
  serialized_end=254,
  methods=[
  _descriptor.MethodDescriptor(
    name='ListEndpoints',
    full_name='Ydb.Discovery.V1.DiscoveryService.ListEndpoints',
    index=0,
    containing_service=None,
    input_type=protos_dot_ydb__discovery__pb2._LISTENDPOINTSREQUEST,
    output_type=protos_dot_ydb__discovery__pb2._LISTENDPOINTSRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='WhoAmI',
    full_name='Ydb.Discovery.V1.DiscoveryService.WhoAmI',
    index=1,
    containing_service=None,
    input_type=protos_dot_ydb__discovery__pb2._WHOAMIREQUEST,
    output_type=protos_dot_ydb__discovery__pb2._WHOAMIRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_DISCOVERYSERVICE)

DESCRIPTOR.services_by_name['DiscoveryService'] = _DISCOVERYSERVICE

# @@protoc_insertion_point(module_scope)



try:
    from sqlalchemy.types import Integer
    from sqlalchemy.sql import type_api
    from sqlalchemy.sql.elements import ColumnElement
    from sqlalchemy import util, exc
except ImportError:
    Integer = object
    ColumnElement = object


class UInt32(Integer):
    __visit_name__ = "uint32"


class UInt64(Integer):
    __visit_name__ = "uint64"


class UInt8(Integer):
    __visit_name__ = "uint8"


class Lambda(ColumnElement):

    __visit_name__ = "lambda"

    def __init__(self, func):
        if not util.callable(func):
            raise exc.ArgumentError("func must be callable")

        self.type = type_api.NULLTYPE
        self.func = func
